from libifstate.util import logger
from libifstate.link.physical import PhysicalLink

class DsaLink(PhysicalLink):
    """
    Distributed Switch Architecture (DSA) user interface
    
    https://docs.kernel.org/networking/dsa/configuration.html
    """
    pass
